"""Export xpypact dataset to pandas dataframes.

This can be used to store the data in databases or parquet datasets.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from xarray import Dataset


def get_run_data(ds: Dataset) -> pd.DataFrame:
    """Extract FISPACT run data as dataframe.

    Args:
        ds: xpypact dataset loaded from FISPACT output JSON file

    Returns:
        pd.DataFrame: one row table with the run data
    """
    return pd.DataFrame.from_dict(
        {
            "timestamp": ds.timestamp,
            "run_name": ds.attrs["run_name"],
            "flux_name": ds.attrs["flux_name"],
            "dose_rate_type": ds.attrs["dose_rate_type"],
            "dose_rate_distance": ds.attrs["dose_rate_distance"],
        },
    )


def get_time_steps(ds: Dataset) -> pd.DataFrame:
    """Extract time steps table from xpypact dataset.

    Drops "total_" prefix from column names.

    Args:
        ds: xpypact dataset to extract from

    Returns:
        Time step totals table.
    """
    return (
        ds[
            [
                "time_step_number",
                "elapsed_time",
                "irradiation_time",
                "cooling_time",
                "duration",
                "flux",
                "total_atoms",
                "total_activity",
                "total_alpha_activity",
                "total_beta_activity",
                "total_gamma_activity",
                "total_mass",
                "total_heat",
                "total_alpha_heat",
                "total_beta_heat",
                "total_gamma_heat",
                "total_ingest1ion_dose",
                "total_inhalation_dose",
                "total_dose_rate",
            ]
        ]
        .to_dataframe()
        .reset_index()
        .rename(
            columns={
                "total_atoms": "atoms",
                "total_activity": "activity",
                "total_alpha_activity": "alpha_activity",
                "total_beta_activity": "beta_activity",
                "total_gamma_activity": "gamma_activity",
                "total_mass": "mass",
                "total_heat": "heat",
                "total_alpha_heat": "alpha_heat",
                "total_beta_heat": "beta_heat",
                "total_gamma_heat": "gamma_heat",
                "total_ingest1ion_dose": "ingest1ion_dose",
                "total_inhalation_dose": "inhalation_dose",
                "total_dose_rate": "dose_rate",
            },
        )
    )


def get_nuclides(ds: Dataset) -> pd.DataFrame:
    """Extract nuclides table from xpypact dataset.

    Sets 'half_life' column name instead 'nuclide_half_life'.

    Args:
        ds: xpypact dataset to extract from

    Returns:
        Nuclide table
    """
    columns_all = ["element", "mass_number", "state", "zai", "nuclide_half_life"]
    columns = ["zai", "nuclide_half_life"]
    return (
        ds[columns]
        .to_dataframe()[columns]
        .reset_index()[columns_all]
        .rename(columns={"nuclide_half_life": "half_life"})
    )


def get_timestep_nuclides(ds: Dataset) -> pd.DataFrame:
    """Extract timestep X nuclides table from xpypact dataset.

    Removes "nuclide_" prefix from column names.

    Args:
        ds: xpypact dataset to extract from

    Returns:
        Time step x Nuclide table
    """
    columns = [
        "nuclide_atoms",
        "nuclide_grams",
        "nuclide_activity",
        "nuclide_alpha_activity",
        "nuclide_beta_activity",
        "nuclide_gamma_activity",
        "nuclide_heat",
        "nuclide_alpha_heat",
        "nuclide_beta_heat",
        "nuclide_gamma_heat",
        "nuclide_dose",
        "nuclide_ingestion",
        "nuclide_inhalation",
    ]
    columns_all = ["time_step_number", "element", "mass_number", "state", *columns]

    return (
        ds[columns]
        .to_dataframe()[columns]
        .reset_index()[columns_all]
        .fillna(0.0)
        .rename(
            columns={
                "nuclide_atoms": "atoms",
                "nuclide_grams": "grams",
                "nuclide_activity": "activity",
                "nuclide_alpha_activity": "alpha_activity",
                "nuclide_beta_activity": "beta_activity",
                "nuclide_gamma_activity": "gamma_activity",
                "nuclide_heat": "heat",
                "nuclide_alpha_heat": "alpha_heat",
                "nuclide_beta_heat": "beta_heat",
                "nuclide_gamma_heat": "gamma_heat",
                "nuclide_dose": "dose",
                "nuclide_ingestion": "ingestion",
                "nuclide_inhalation": "inhalation",
            },
        )
    )


def get_gamma(ds: Dataset) -> pd.DataFrame | None:
    """Extract timestep X gamma table from xpypact dataset.

    Set column names to "boundary" and "rate".

    Args:
        ds: xpypact dataset to extract from

    Returns:
        Time step x gamma rate table
    """
    if "gamma" not in ds:
        return None
    columns = [
        "time_step_number",
        "gamma_boundaries",
        "gamma",
    ]
    return (
        ds.gamma.to_dataframe()
        .reset_index()[columns]
        .rename(columns={"gamma_boundaries": "boundary", "gamma": "rate"})
    )



"""Utility types for the package."""
from __future__ import annotations

from typing import Union

import os

import numpy as np

from numpy.typing import NDArray

MayBePath = Union[str, os.PathLike, None]
NDArrayFloat = NDArray[np.float_]
NDArrayInt = NDArray[np.int_]



"""Classes to load information from FISPACT output JSON file."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterator, cast

import io  # noqa: TCH003 - needed for dispatch

from dataclasses import dataclass
from functools import singledispatch
from pathlib import Path  # noqa: TCH003 - needed for dispatch

import numpy as np

import orjson as json

from xpypact.run_data import RunData
from xpypact.time_step import TimeStep

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from xpypact.utils.types import NDArrayFloat

FLOAT_ZERO = 0.0


@dataclass
class RunDataCorrected:
    """Common data for an FISPACT inventory.

    Note:
        Correction - dose_rate_type and dose_rate_distance are duplicated
        the FISPACT time steps. This information is extracted to this header.
    """

    timestamp: str
    run_name: str
    flux_name: str
    dose_rate_type: str
    dose_rate_distance: float


class InventoryError(ValueError):
    """Base class for inventory exception."""

    def __str__(self) -> str:
        """Use the class __doc__ as a message.

        Returns:
            The __doc__ of the exception class.
        """
        return cast(str, self.__class__.__doc__)  # pragma: no cover


def _create_json_inventory_data_mapper() -> Callable[[dict[str, Any]], TimeStep]:
    prev_irradiation_time = prev_cooling_time = prev_elapsed_time = FLOAT_ZERO
    number = 1

    def json_inventory_data_mapper(jts: dict[str, Any]) -> TimeStep:
        nonlocal number, prev_irradiation_time, prev_cooling_time, prev_elapsed_time
        ts = TimeStep.from_json(jts)
        duration = ts.irradiation_time - prev_irradiation_time
        if duration == FLOAT_ZERO:
            duration = ts.cooling_time - prev_cooling_time
        if duration < FLOAT_ZERO:
            raise InventoryNonMonotonicTimesError  # pragma: no cover
        ts.duration = duration
        prev_elapsed_time = ts.elapsed_time = prev_elapsed_time + duration
        if duration == FLOAT_ZERO:
            ts.flux = FLOAT_ZERO
        ts.number = number
        number += 1
        prev_irradiation_time, prev_cooling_time = (
            ts.irradiation_time,
            ts.cooling_time,
        )
        return ts

    return json_inventory_data_mapper


class InventoryNonMonotonicTimesError(InventoryError):
    """Irradiation and cooling times in FISPACT output should be monotonically increasing."""


@dataclass
class Inventory:
    """Helper class to load FISPACT inventory (output) from JSON file.

    Implements list interface over time steps.
    """

    run_data: RunData
    inventory_data: list[TimeStep]

    @property
    def meta_info(self) -> RunDataCorrected:
        """Create corrected Inventory header.

        Returns:
            RunDataCorrected: with common data for the inventory.
        """
        ts = self.inventory_data[-1]  # The dose_rate in the first timestep can be empty.
        return RunDataCorrected(
            self.run_data.timestamp,
            self.run_data.run_name,
            self.run_data.flux_name,
            ts.dose_rate.type,
            ts.dose_rate.distance,
        )

    @classmethod
    def from_json(cls, json_dict: dict[str, Any]) -> Inventory:
        """Construct Inventory instance from JSON dictionary.

        Args:
            json_dict: a JSON dictionary

        Returns:
            Inventory: the new instance
        """
        json_run_data = json_dict.pop("run_data")
        run_data = RunData.from_json(json_run_data)
        json_inventory_data = json_dict.pop("inventory_data")
        mapper = _create_json_inventory_data_mapper()
        inventory_data = [mapper(ts) for ts in json_inventory_data]

        return cls(run_data, inventory_data)

    def extract_times(self) -> NDArrayFloat:
        """Create vector of elapsed time for all the time steps in the inventory.

        Returns:
            Vector with elapsed times.
        """
        return extract_times(self.inventory_data)

    def __iter__(self) -> Iterator[TimeStep]:
        """Iterate over time steps.

        Returns:
            Iterator over the time steps.
        """
        return iter(self.inventory_data)

    def __len__(self) -> int:
        """Length, delegated to time steps.

        Returns:
            int: length of the time steps list.
        """
        return len(self.inventory_data)

    def __getitem__(self, item):
        """List interface delegated to the time steps.

        Args:
            item: an item index or slice

        Returns:
            Selected item or items.
        """
        return self.inventory_data[item]


def extract_times(time_steps: Iterable[TimeStep]) -> NDArrayFloat:
    """Create vector of elapsed time for all the time steps.

    Args:
        time_steps: list of time steps

    Returns:
        Vector with elapsed times.
    """
    return np.fromiter((x.elapsed_time for x in time_steps), dtype=float)


# @dispatch(str)
@singledispatch
def from_json(text: str) -> Inventory:
    """Construct Inventory instance from JSON text.

    Args:
        text: Text source.

    Returns:
        The loaded Inventory instance.
    """
    json_dict = json.loads(text)  # pylint: disable=no-member
    return Inventory.from_json(json_dict)


@from_json.register
def _(stream: io.IOBase) -> Inventory:  # type: ignore[misc]
    """Construct Inventory instance from JSON stream.

    Args:
        stream: IO stream source.

    Returns:
        The loaded Inventory instance.
    """
    return from_json(stream.read())


@from_json.register
def _(path: Path) -> Inventory:  # type: ignore[misc]
    """Construct Inventory instance from JSON path.

    Args:
        path: path to source.

    Returns:
        The loaded Inventory instance.
    """
    return from_json(path.read_text(encoding="utf8"))



"""Output utilities."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, TextIO

import sys

if TYPE_CHECKING:
    from collections.abc import Iterable


def print_cols(
    seq: Iterable[Any],
    fid: TextIO = sys.stdout,
    max_columns: int = 6,
    fmt: str = "{}",
) -> int:
    """Print sequence in columns.

    Args:
        seq: sequence to print
        fid: output
        max_columns: max columns in a line
        fmt: format string

    Returns:
        int: the number of the last column printed on the last row
    """
    i = 0
    for s in seq:
        print(fmt.format(s), file=fid, end="")
        if (i > 0) and (i % max_columns == 0):
            print(file=fid)
        else:
            print(" ", file=fid, end="")
        i += 1
    return i % max_columns



"""Utility methods to access a package data."""
from __future__ import annotations

from typing import TYPE_CHECKING

from importlib.resources import files

if TYPE_CHECKING:
    from collections.abc import Callable
    from importlib.abc import Traversable
    from importlib.resources import Package
    from pathlib import Path


def path_resolver(package: Package) -> Callable[[str], Path | Traversable]:
    """Create method to find data path.

    Args:
        package: the package below which the data is stored.

    Returns:
        callable which appends the argument to the package folder as Path.
    """
    return files(package).joinpath



"""Methods to convert Inventory to xarray DataSet.

The dimensions  ot the nodes are:
    time_step_number
    nuclide (symbol, atomic_number, state)
    gamma_boundaries if available

The dataset facilitates FISPACT output slicing and aggregation.

Also, the module provides methods to scale activation properties in
activation and time_steps nodes are provided, both for mass scaling
(all the values proportional to mass are multiplied by scaling factor)
and for flux scaling:
the difference with initial state is scaled with a given factor.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TextIO, cast

from functools import reduce

import numpy as np

from numpy.typing import ArrayLike

import pandas as pd
import xarray as xr

from mckit_nuclides.elements import ELEMENTS_TABLE
from mckit_nuclides.nuclides import NUCLIDES_TABLE
from xarray.core.accessor_dt import DatetimeAccessor
from xpypact.inventory import from_json as inventory_from_json

if TYPE_CHECKING:  # pragma: no cover
    from pathlib import Path

    try:
        from dask.delayed import Delayed
    except ImportError:
        Delayed = None
    try:
        from dask.dataframe import DataFrame as DaskDataFrame
    except ImportError:
        DaskDataFrame = None

    from xpypact.inventory import Inventory
    from xpypact.time_step import TimeStep
    from xpypact.utils.types import MayBePath

SCALABLE_COLUMNS = [
    "total_mass",
    "total_heat",
    "total_dose_rate",
    "total_alpha_heat",
    "total_beta_heat",
    "total_gamma_heat",
    "total_activity",
    "total_alpha_activity",
    "total_beta_activity",
    "total_gamma_activity",
    "nuclide_atoms",
    "nuclide_grams",
    "nuclide_activity",
    "nuclide_alpha_activity",
    "nuclide_beta_activity",
    "nuclide_gamma_activity",
    "nuclide_heat",
    "nuclide_alpha_heat",
    "nuclide_beta_heat",
    "nuclide_gamma_heat",
    "gamma",
]


def _make_var(value, dtype=float) -> tuple[str, ArrayLike]:
    value = np.array([value], dtype=dtype)
    return "time_step_number", value


def _make_nuclide_var(getter, nuclides, dtype=float) -> tuple[str, ArrayLike]:
    return "nuclide", np.fromiter(map(getter, nuclides), dtype=dtype)


def _make_time_step_and_nuclide_var(
    getter,
    nuclides,
    dtype=float,
) -> tuple[tuple[str, str], ArrayLike]:
    _data = np.fromiter(map(getter, nuclides), dtype=dtype)
    return ("time_step_number", "nuclide"), _data.reshape(1, _data.size)


def _add_time_step_record(_ds: xr.Dataset, ts: TimeStep) -> xr.Dataset:
    data_vars = {
        "irradiation_time": _make_var(ts.irradiation_time),
        "cooling_time": _make_var(ts.cooling_time),
        "duration": _make_var(ts.duration),
        "flux": _make_var(ts.flux),
        "total_atoms": _make_var(ts.total_atoms),
        "total_activity": _make_var(ts.total_activity),
        "total_alpha_activity": _make_var(ts.alpha_activity),
        "total_beta_activity": _make_var(ts.beta_activity),
        "total_gamma_activity": _make_var(ts.gamma_activity),
        "total_mass": _make_var(ts.total_mass),
        "total_heat": _make_var(ts.total_heat),
        "total_alpha_heat": _make_var(ts.alpha_heat),
        "total_beta_heat": _make_var(ts.beta_heat),
        "total_gamma_heat": _make_var(ts.gamma_heat),
        "total_ingest1ion_dose": _make_var(ts.ingestion_dose),
        "total_inhalation_dose": _make_var(ts.inhalation_dose),
        "total_dose_rate": _make_var(ts.dose_rate.dose),
        "nuclide_half_life": _make_nuclide_var(lambda n: n.half_life, ts.nuclides),
        "nuclide_atoms": _make_time_step_and_nuclide_var(lambda n: n.atoms, ts.nuclides),
        "nuclide_grams": _make_time_step_and_nuclide_var(lambda n: n.grams, ts.nuclides),
        "nuclide_activity": _make_time_step_and_nuclide_var(lambda n: n.activity, ts.nuclides),
        "nuclide_alpha_activity": _make_time_step_and_nuclide_var(
            lambda n: n.alpha_activity,
            ts.nuclides,
        ),
        "nuclide_beta_activity": _make_time_step_and_nuclide_var(
            lambda n: n.beta_activity,
            ts.nuclides,
        ),
        "nuclide_gamma_activity": _make_time_step_and_nuclide_var(
            lambda n: n.gamma_activity,
            ts.nuclides,
        ),
        "nuclide_heat": _make_time_step_and_nuclide_var(lambda n: n.heat, ts.nuclides),
        "nuclide_alpha_heat": _make_time_step_and_nuclide_var(lambda n: n.alpha_heat, ts.nuclides),
        "nuclide_beta_heat": _make_time_step_and_nuclide_var(lambda n: n.beta_heat, ts.nuclides),
        "nuclide_gamma_heat": _make_time_step_and_nuclide_var(lambda n: n.gamma_heat, ts.nuclides),
        "nuclide_dose": _make_time_step_and_nuclide_var(lambda n: n.dose, ts.nuclides),
        "nuclide_ingestion": _make_time_step_and_nuclide_var(lambda n: n.ingestion, ts.nuclides),
        "nuclide_inhalation": _make_time_step_and_nuclide_var(lambda n: n.inhalation, ts.nuclides),
    }

    coords: dict[str, Any] = {
        "time_step_number": _make_var(ts.number, dtype=int),
        "elapsed_time": _make_var(ts.elapsed_time),
        "zai": _make_nuclide_var(lambda n: n.zai, ts.nuclides, dtype=int),
    }

    if ts.gamma_spectrum is not None:
        gamma_boundaries = ts.gamma_spectrum.boundaries
        gamma_value = ts.gamma_spectrum.intensities
        gamma_value = np.insert(gamma_value, 0, 0.0).reshape(1, len(gamma_boundaries))
        data_vars["gamma"] = (("time_step_number", "gamma_boundaries"), gamma_value)
        coords["gamma_boundaries"] = gamma_boundaries

    tsr = xr.Dataset(data_vars=data_vars, coords=coords)

    nuclide_coordinate = pd.MultiIndex.from_tuples(
        ((n.element, n.a, n.state) for n in ts.nuclides),
        names=["element", "mass_number", "state"],
    )

    nuclide_coordinate_idx = xr.Coordinates.from_pandas_multiindex(
        nuclide_coordinate,
        dim="nuclide",
    )

    tsr = tsr.assign_coords(nuclide_coordinate_idx)
    return xr.merge([_ds, tsr])


def create_dataset(inventory: Inventory) -> xr.Dataset:
    """Convert Inventory to a Dataset.

    Args:
        inventory: source

    Returns:
        Dataset: the representation of a FISPACT inventory.
    """
    ds = reduce(_add_time_step_record, iter(inventory), xr.Dataset())
    meta_info = inventory.meta_info
    ds.coords["timestamp"] = pd.date_range(meta_info.timestamp, periods=1)
    ds.coords["timestamp"].attrs[
        "long description"
    ] = "FISPACT datasets can be merged and then selected by timestamp as coordinate"
    ds.attrs["run_name"] = meta_info.run_name
    ds.attrs["flux_name"] = meta_info.flux_name
    ds.attrs["dose_rate_type"] = meta_info.dose_rate_type
    ds.attrs["dose_rate_distance"] = meta_info.dose_rate_distance
    ds.irradiation_time.attrs["units"] = "s"
    ds.cooling_time.attrs["units"] = "s"
    ds.duration.attrs["units"] = "s"
    ds.elapsed_time.attrs["units"] = "s"
    ds.flux.attrs["units"] = "n/cm^2/s"

    if "gamma_boundaries" in ds.coords:
        ds.coords["gamma_boundaries"].attrs["units"] = "MeV"

    ds.total_activity.attrs["units"] = "Bq"
    ds.total_alpha_activity.attrs["units"] = "Bq"
    ds.total_beta_activity.attrs["units"] = "Bq"
    ds.total_gamma_activity.attrs["units"] = "Bq"

    ds.total_mass.attrs["units"] = "kg"

    ds.total_heat.attrs["units"] = "kW"
    ds.total_alpha_heat.attrs["units"] = "kW"
    ds.total_beta_heat.attrs["units"] = "kW"
    ds.total_gamma_heat.attrs["units"] = "kW"

    ds.total_ingest1ion_dose.attrs["units"] = "Sv"
    ds.total_inhalation_dose.attrs["units"] = "Sv"
    ds.total_dose_rate.attrs["units"] = "Sv/h"

    ds.nuclide_half_life.attrs["units"] = "s"
    ds.nuclide_grams.attrs["units"] = "g"

    ds.nuclide_activity.attrs["units"] = "Bq"
    ds.nuclide_alpha_activity.attrs["units"] = "Bq"
    ds.nuclide_beta_activity.attrs["units"] = "Bq"
    ds.nuclide_gamma_activity.attrs["units"] = "Bq"

    ds.nuclide_heat.attrs["units"] = "kW"
    ds.nuclide_alpha_heat.attrs["units"] = "kW"
    ds.nuclide_beta_heat.attrs["units"] = "kW"
    ds.nuclide_gamma_heat.attrs["units"] = "kW"

    ds.nuclide_dose.attrs["units"] = "Sv"
    ds.nuclide_ingestion.attrs["units"] = "Sv"
    ds.nuclide_inhalation.attrs["units"] = "Sv"

    # Restore int type for 'zai', which became float after the merges.
    # see https://xarray.pydata.org/en/stable/user-guide/combining.html
    # ...  due  to the underlying representation of
    # missing values as floating point numbers(NaN),
    # variable  data type is not always  preserved when merging in this manner.
    zais = np.asarray(ds.coords["zai"].to_numpy().flatten(), dtype=int)
    ds.coords["zai"] = ("nuclide", zais)

    return ds


def get_timestamp(ds: xr.Dataset) -> DatetimeAccessor:
    """Get access to time stamp value.

    Args:
        ds: Dataset with FISPACT results

    Returns:
        accessor to timestamp
    """
    return cast(DatetimeAccessor, ds.timestamp[0].dt)


def get_atomic_numbers(ds: xr.Dataset) -> ArrayLike:
    """Create column of atomic numbers (Z) corresponding to `nuclide` coordinate.

    Args:
        ds: dataset with the 'nuclides' coordinate

    Returns:
        Z values for the nuclides
    """
    elements = ds.nuclide.element.to_numpy()
    return cast(
        ArrayLike,
        ELEMENTS_TABLE.loc[elements, ["atomic_number"]].to_numpy().flatten(),
    )


def add_atomic_number_coordinate(
    ds: xr.Dataset,
    coordinate_name: str = "atomic_number",
) -> xr.Dataset:
    """Add coordinate for Z.

    This allows indexing of a dataset with integer Z (atomic number) values.

    Args:
        ds: FISPACT results dataset
        coordinate_name: name for the coordinate attribute

    Returns:
        The 'ds' with the added coordinate.
    """
    ds.coords[coordinate_name] = ("nuclide", get_atomic_numbers(ds))
    return ds


def get_atomic_masses(ds: xr.Dataset) -> ArrayLike:
    """Create column with relative atomic masses along nuclide coordinate.

    Args:
        ds: FISPACT output as Dataset

    Returns:
        Vector with relative atomic masses.
    """
    atomic_numbers = get_atomic_numbers(ds)
    mass_numbers = ds.nuclide.mass_number.to_numpy().flatten()
    mi = pd.MultiIndex.from_arrays([atomic_numbers, mass_numbers])
    return cast(ArrayLike, NUCLIDES_TABLE.loc[mi, ["nuclide_mass"]].to_numpy().flatten())


def add_atomic_masses(ds: xr.Dataset, variable_name="atomic_mass") -> xr.Dataset:
    """Add variable with relative atomic masses with 'nuclide' coordinate.

    Useful for bulk operations based on coordinates.

    Args:
        ds: FISPACT output as Dataset
        variable_name: the data array name to use

    Returns:
        The 'ds' with added data array with relative atomic masses.
    """
    ds[variable_name] = ("nuclide", get_atomic_masses(ds))
    return ds


def scale_by_flux(ds: xr.Dataset, scale: float) -> xr.Dataset:
    """Normalize irradiation results for actual flux value.

    A FISPACT irradiation scenario may use flux value,
    which is different from actual one.
    This is the case for standard ITER SA2 scenario.
    For flux, only difference with initial activation properties is to be scaled.

    Notes:
        Assuming that the first time step presents initial values.

    Args:
        ds: FISPACT output as Dataset
        scale: factor to apply

    Returns:
        A new dataset with scaled columns.
    """
    initial = ds.sel(time_step_number=1).fillna(0.0)
    result = scale_by_mass(ds - initial, scale) + initial
    result.attrs.update(ds.attrs)

    return result


def scale_by_mass(ds: xr.Dataset, scale: float) -> xr.Dataset:
    """All the activation values including initial can be scaled to actual weight.

    Args:
        ds: FISPACT output as Dataset
        scale: factor to apply

    Returns:
        A new dataset with scaled columns.
    """
    columns = [x for x in SCALABLE_COLUMNS if x in ds.variables]
    return ds.merge(ds[columns] * scale, overwrite_vars=columns, join="exact")


def _encode_multiindex(ds: xr.Dataset, idx_name: str) -> xr.Dataset:
    coordinate = ds.indexes[idx_name].to_frame().reset_index(drop=True)
    columns = coordinate.columns.to_numpy()
    columns_to_drop = (idx_name, *columns)
    encoded = ds.reset_index(columns_to_drop, drop=True)
    encoded.attrs[idx_name + "_columns"] = columns
    for c in columns:
        encoded.attrs[idx_name + f"_{c}"] = coordinate[c].to_numpy().flatten()
    return encoded


def _decode_to_multiindex(encoded: xr.Dataset, idx_name: str) -> xr.Dataset:
    columns = encoded.attrs.pop(idx_name + "_columns")
    mi_columns = [encoded.attrs.pop(idx_name + f"_{c}") for c in columns]
    mi = pd.MultiIndex.from_arrays(mi_columns, names=columns)
    mi_idx = xr.Coordinates.from_pandas_multiindex(mi, dim=idx_name)
    return encoded.assign_coords(mi_idx)


# noinspection PyShadowingBuiltins
def save_nc(
    ds: xr.Dataset,
    path: MayBePath = None,
    **kwargs: Any,
) -> bytes | Delayed | None:
    """Save a dataset with nuclide index.

    Encodes MultiIndex instance in a dataset and saves the result.
    See: https://github.com/pydata/xarray/issues/1077

    Args:
        ds: FISPACT output as dataset
        path: see :meth:`xarray.Dataset.to_netcdf()`
        kwargs: ...

    Returns:
        see :meth:`xarray.Dataset.to_netcdf()`
    """
    encoded = _encode_multiindex(ds, "nuclide")
    engine = kwargs.pop("engine", "h5netcdf")
    return encoded.to_netcdf(path, engine=engine, **kwargs)


def load_nc(
    cn: str | Path,
    engine: Literal["netcdf4", "scipy", "h5netcdf"] | None = "h5netcdf",
    **kwargs: Any,
) -> xr.Dataset:
    """Load a dataset from a netcdf file.

    The nuclide index is restored.

    Args:
        cn: path to dataset file
        engine: netcdf engine to use
        kwargs: arguments for :meth:`xarray.Dataset.load_dataset()`

    Returns:
        The loaded dataset
    """
    encoded = xr.load_dataset(cn, engine=engine, **kwargs)
    return _decode_to_multiindex(encoded, "nuclide")


def from_json(path: str | Path | TextIO) -> xr.Dataset:
    """Load Dataset from the FISPACT JSON output.

    Args:
        path: path to JSON file, stream or text with FISPACT JSON output.

    Returns:
        The loaded Dataset.
    """
    return create_dataset(inventory_from_json(path))  # type: ignore[arg-type]



"""Code to implement DuckDB DAO."""
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from dataclasses import dataclass
from multiprocessing import cpu_count
from pathlib import Path

import duckdb as db

from xpypact import get_gamma, get_nuclides, get_run_data, get_time_steps, get_timestep_nuclides
from xpypact.dao import DataAccessInterface
from xpypact.utils.resource import path_resolver

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr


# On using DuckDB with multiple threads, see
# https://duckdb.org/docs/guides/python/multiple_threads.html


# noinspection SqlNoDataSourceInspection
@dataclass
class DuckDBDAO(DataAccessInterface):
    """Implementation of DataAccessInterface for DuckDB."""

    con: db.DuckDBPyConnection

    def get_tables_info(self) -> pd.DataFrame:
        """Get information on tables in schema."""
        return self.con.execute("select * from information_schema.tables").df()

    @property
    def tables(self) -> tuple[str, str, str, str, str]:
        """List tables being used by xpypact dao.

        Returns:
            Tuple with table names.
        """
        return "rundata", "timestep", "nuclide", "timestep_nuclide", "timestep_gamma"

    def has_schema(self) -> bool:
        """Check if the schema is available in a database."""
        db_tables = self.get_tables_info()

        if len(db_tables) < len(self.tables):
            return False

        table_names = db_tables["table_name"].to_numpy()

        return all(name in table_names for name in self.tables)

    def create_schema(self) -> None:
        """Create tables to store xpypact dataset.

        Retain existing tables.
        """
        sql_path: Path = cast(Path, path_resolver(__package__)("create_schema.sql"))
        sql = sql_path.read_text(encoding="utf-8")
        self.con.execute(sql)

    def drop_schema(self) -> None:
        """Drop our DB objects."""
        tables = [
            "timestep_nuclide",
            "timestep_gamma",
            "timestep",
            "nuclide",
            "rundata",
        ]
        for table in tables:
            self.con.execute(f"drop table if exists {table}")

    def save(self, ds: xr.Dataset, material_id=1, case_id=1) -> None:
        """Save xpypact dataset to database.

        Args:
            ds: xpypact dataset to save
            material_id: additional key to distinguish multiple FISPACT run
            case_id: second additional key
        """
        self._save_run_data(ds, material_id, case_id)
        self._save_nuclides(ds)
        self._save_time_steps(ds)
        self._save_time_step_nuclides(ds)
        self._save_gamma(ds)

    def load_rundata(self) -> db.DuckDBPyRelation:
        """Load FISPACT run data as table.

        Returns:
            FISPACT run data ad RelObj
        """
        return self.con.table("rundata")

    def load_nuclides(self) -> db.DuckDBPyRelation:
        """Load nuclide table.

        Returns:
            time nuclide
        """
        return self.con.table("nuclide")

    def load_time_steps(self) -> db.DuckDBPyRelation:
        """Load time step table.

        Returns:
            time step table
        """
        return self.con.table("timestep")

    def load_time_step_nuclides(self) -> db.DuckDBPyRelation:
        """Load time step x nuclides table.

        Returns:
            time step x nuclides table
        """
        return self.con.table("timestep_nuclide")

    def load_gamma(self, time_step_number: int | None = None) -> db.DuckDBPyRelation:
        """Load time step x gamma table.

        Args:
            time_step_number: filter for time_step_number

        Returns:
            time step x gamma table
        """
        sql = "select * from timestep_gamma"
        if time_step_number is not None:
            sql += f" where time_step_number == {time_step_number}"
        return self.con.sql(sql)

    def _save_run_data(self, ds: xr.Dataset, material_id=1, case_id=1) -> None:
        _table = get_run_data(ds)
        _table = _add_material_and_case_columns(_table, material_id, case_id)
        sql = """
            begin transaction;
            insert into rundata from _table;
            commit;
            """
        self.con.sql(sql)

    def _save_nuclides(self, ds: xr.Dataset):
        _table = get_nuclides(ds)  # noqa: F841 - used below
        sql = """
            begin transaction;
            insert or ignore into nuclide select * from _table;
            commit;
        """
        self.con.sql(sql)

    def _save_time_steps(self, ds: xr.Dataset, material_id=1, case_id=1):
        _table = get_time_steps(ds)
        _table = _add_material_and_case_columns(_table, material_id, case_id)
        sql = """
            begin transaction;
            insert into timestep from _table;
            commit;
            """
        self.con.execute(sql)
        self.con.commit()

    def _save_time_step_nuclides(self, ds: xr.Dataset, material_id=1, case_id=1):
        _table = get_timestep_nuclides(ds)
        _table = _add_material_and_case_columns(_table, material_id, case_id)
        sql = """
            begin transaction;
            insert into timestep_nuclide from _table;
            commit;
            """
        self.con.sql(sql)

    def _save_gamma(self, ds: xr.Dataset, material_id=1, case_id=1):
        _table = get_gamma(ds)
        _table = _add_material_and_case_columns(_table, material_id, case_id)
        self.con.sql(
            """
            begin transaction;
            insert into timestep_gamma from _table;
            commit;
            """,
        )


def compute_optimal_row_group_size(frame_size: int, _cpu_count: int = 0) -> int:
    """Compute DuckDB row group size for writing to parquet files.

    This should optimize speed of reading from parquet files.

    Args:
        frame_size: length of rows to be written
        _cpu_count: number to split the rows

    Returns:
        the row group size
    """
    if not _cpu_count:
        _cpu_count = max(4, cpu_count())
    size = frame_size // _cpu_count
    return min(max(size, 2048), 1_000_000)


def write_parquet(target_dir: Path, ds: xr.Dataset, material_id: int, case_id: int) -> None:
    """Store xpypact dataset to parquet directories.

    Create in 4 subdirectories in `target_dir` for data subjects run_data, time_steps,
    time_step_nuclides, and gamma dataframes.
    Save the dataframes as parquet files.
    Hive partiotioning is not used in this version, because resulting partions are too small.
    The data for  can be  saved in the same `target_dir` as long as the material_id and case_id are
    unique for an data subject.

    This structure can be easily and efficiently accessed from DuckDB as external data.

    For instance to collect all the inventories::

        select * from read_parquet('<target_dir>/inventory/*/*/*.parquet', hive_partitioning=true)

    We use in memory DuckDB instance to transfer data to parquet to ensure compatibility
    for later reading back to DuckDB.

    See about row_group_size: https://duckdb.org/docs/data/parquet/tips

    Args:
        target_dir: root directory to store a dataset in subdirectories
        ds: dataset to store
        material_id: the level 1 key to distinguish with other datasets in the folder,
                     may correspond to material in R2S
        case_id: the level 2 key, may correspond neutron group or case in R2S of fispact run.
                 The case may be registered in the database to provide additional information.
    """
    to_proces: dict[str, pd.DataFrame] = {
        "run_data": get_run_data(ds),
        "nuclides": get_nuclides(ds),
        "time_steps": get_time_steps(ds),
        "timestep_nuclides": get_timestep_nuclides(ds),
        "gamma": get_gamma(ds),
    }
    con = db.connect()
    try:
        for k, v in to_proces.items():
            path: Path = target_dir / k
            path.mkdir(parents=True, exist_ok=True)
            frame = _add_material_and_case_columns(
                v,
                material_id,
                case_id,
            )
            sql = f"""
                copy
                (select * from frame)
                to '{path}'
                (
                    format parquet,
                    per_thread_output true,
                    overwrite_or_ignore true,
                    filename_pattern "data_material_id={material_id}_case_id={case_id}_{{i}}",
                    row_group_size {compute_optimal_row_group_size(frame.shape[0])}
                )
                """  # noqa: S608 - sql injection
            con.execute(sql)
    finally:
        con.close()


def _add_material_and_case_columns(
    table: pd.DataFrame,
    material_id: int,
    case_id: int,
) -> pd.DataFrame:
    columns = table.columns.to_numpy()
    table["material_id"] = material_id
    table["case_id"] = case_id
    new_columns = ["material_id", "case_id", *columns]
    return table[new_columns]



"""The Class to represent FISPACT fluxes file content."""
from __future__ import annotations

from typing import TYPE_CHECKING, TextIO, cast

from dataclasses import dataclass
from io import StringIO, TextIOBase
from pathlib import Path

import numpy as np

from numpy import allclose, array_equal

from multipledispatch import dispatch
from xpypact.utils.io import print_cols

FISPACT_709_BINS_NUMBER = 709

if TYPE_CHECKING:
    from collections.abc import Callable

    from xpypact.utils.types import NDArrayFloat

# pylint: disable=function-redefined

LAST_TWO_DECADES = np.fromstring(
    """
        1.0000E+9
        9.6000E+8
        9.2000E+8
        8.8000E+8
        8.4000E+8
        8.0000E+8
        7.6000E+8
        7.2000E+8
        6.8000E+8
        6.4000E+8
        6.0000E+8
        5.6000E+8
        5.2000E+8
        4.8000E+8
        4.4000E+8
        4.0000E+8
        3.6000E+8
        3.2000E+8
        2.8000E+8
        2.4000E+8
        2.0000E+8
        1.8000E+8
        1.6000E+8
        1.5000E+8
        1.4000E+8
        1.3000E+8
        1.2000E+8
        1.1000E+8
        1.0000E+8
        9.0000E+7
        8.0000E+7
        7.5000E+7
        7.0000E+7
        6.5000E+7
        6.0000E+7
        5.8000E+7
        5.6000E+7
        5.4000E+7
        5.2000E+7
        5.0000E+7
        4.8000E+7
        4.6000E+7
        4.4000E+7
        4.2000E+7
        4.0000E+7
        3.8000E+7
        3.6000E+7
        3.4000E+7
        3.2000E+7
        3.0000E+7
        2.9000E+7
        2.8000E+7
        2.7000E+7
        2.6000E+7
        2.5000E+7
        2.4000E+7
        2.3000E+7
        2.2000E+7
        2.1000E+7
        2.0000E+7
        1.9800E+7
        1.9600E+7
        1.9400E+7
        1.9200E+7
        1.9000E+7
        1.8800E+7
        1.8600E+7
        1.8400E+7
        1.8200E+7
        1.8000E+7
        1.7800E+7
        1.7600E+7
        1.7400E+7
        1.7200E+7
        1.7000E+7
        1.6800E+7
        1.6600E+7
        1.6400E+7
        1.6200E+7
        1.6000E+7
        1.5800E+7
        1.5600E+7
        1.5400E+7
        1.5200E+7
        1.5000E+7
        1.4800E+7
        1.4600E+7
        1.4400E+7
        1.4200E+7
        1.4000E+7
        1.3800E+7
        1.3600E+7
        1.3400E+7
        1.3200E+7
        1.3000E+7
        1.2800E+7
        1.2600E+7
        1.2400E+7
        1.2200E+7
        1.2000E+7
        1.1800E+7
        1.1600E+7
        1.1400E+7
        1.1200E+7
        1.1000E+7
        1.0800E+7
        1.0600E+7
        1.0400E+7
        1.0200E+7
    """,
    dtype=float,
    sep=" \n",
)[::-1]


def compute_709_bins() -> NDArrayFloat:
    """Computes bins for 709 groups according to the specification.

    Decades from -5 to 9 are divided to 50 bins equidistant in log scale.

    Note:
        The computed bins show better correspondence with FISPACT converted fluxes,
        then available tabulated data.
        See notebooks/dvp/demo_fispact_flux_generator.ipynb

    Returns:
        array of 709 bins
    """
    template = np.logspace(0, 1, 51)
    res = np.hstack([template[1:] * 10**i for i in range(-5, 7)] + [LAST_TWO_DECADES])
    return np.insert(res, 0, 1e-5)


FISPACT_709_BINS = compute_709_bins()


@dataclass(eq=False, order=False)
class Fluxes:
    """Class to represent flux data."""

    energy_bins: NDArrayFloat
    fluxes: NDArrayFloat
    comment: str
    norm: float = 1.0

    def __post_init__(self) -> None:
        """Validate sizes of arrays.

        Raises:
            ValueError: if sizes of bins and fluxes are not compatible.
        """
        if self.fluxes.size != self.energy_bins.size - 1:
            msg = (
                "Incompatible sizes of bins and fluxes, "
                f"{self.fluxes.size} and {self.energy_bins.size}",
            )
            raise ValueError(msg)

    @property
    def total(self) -> float:
        """Return flux total.

        Returns:
            float: flux total
        """
        return cast(float, np.sum(self.fluxes))

    def __hash__(self) -> int:
        """Use available information for hash.

        Returns:
            hash
        """
        return hash((self.energy_bins.size, self.fluxes.size, self.comment, self.norm))

    def __eq__(self, other) -> bool:
        """Compare fluxes.

        Args:
            other: other fluxes

        Returns:
            bool: True, if the fluxes are equal, otherwise - False
        """
        return self is other or (
            isinstance(other, Fluxes)
            and array_equal(self.energy_bins, other.energy_bins)
            and array_equal(self.fluxes, other.fluxes)
            and self.comment == other.comment
            and self.norm == other.norm
        )

    def __ne__(self, other) -> bool:
        """Compare fluxes.

        Args:
            other: other fluxes

        Returns:
            bool: False, if the fluxes are equal, otherwise - True
        """
        return not self.__eq__(other)


def is_709_fluxes(fluxes: Fluxes) -> bool:
    """Check if fluxes are 709-kind of fluxes.

    This fluxes can be output without energy bins and
    can be used with 'flux' statement in FISPACT run configuration.

    Args:
        fluxes: what to check

    Returns:
        bool: True, if fluxes are 709 kind of fluxes.
    """
    return fluxes.energy_bins is FISPACT_709_BINS and fluxes.fluxes.size == FISPACT_709_BINS_NUMBER


def are_fluxes_equal(a: Fluxes, b: Fluxes) -> bool:
    """Compare data of fluxes.

    Check if fluxes are equivalent by data, disregarding differences in comment and norm.

    Args:
        a: the first flux to compare
        b: the second flux to compare

    Returns:
        bool: True, if the fluxes are equivalent by data
    """
    return (
        a.energy_bins.size == b.energy_bins.size
        and array_equal(a.energy_bins, b.energy_bins)
        and array_equal(a.fluxes, b.fluxes)
    )


def are_fluxes_close(
    a: Fluxes,
    b: Fluxes,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
    *,
    equal_nan: bool = False,
) -> bool:
    """Compare data of fluxes approximately.

    Check if fluxes are equivalent by data with some tolerance,
    disregarding differences in comment and norm.

    Args:
        a: the first flux to compare
        b: the second flux to compare
        rtol: relative tolerance
        atol: absolute tolerance
        equal_nan: see :func:`numpy.allclose()`

    Returns:
        bool: True, if the fluxes are equivalent by data
    """
    return (
        a.energy_bins.size == b.energy_bins.size
        and allclose(a.energy_bins, b.energy_bins, rtol=rtol, atol=atol, equal_nan=equal_nan)
        and allclose(a.fluxes, b.fluxes, rtol=rtol, atol=atol, equal_nan=equal_nan)
    )


def read_fluxes(
    stream: TextIO,
    define_bins_and_fluxes: Callable[[NDArrayFloat], tuple[NDArrayFloat, NDArrayFloat]],
) -> Fluxes:
    """Load Fluxes from a stream.

    Args:
        stream: IO to load from
        define_bins_and_fluxes: strategy to extract bins and values

    Returns:
        Fluxes: the loaded
    """
    lines = stream.readlines()
    comment = lines[-1].strip()
    norm = float(lines[-2])
    lines = lines[:-2]
    data = np.fromstring(" ".join(lines), dtype=float, count=-1, sep=" ")
    energy_bins, fluxes = define_bins_and_fluxes(data)
    return Fluxes(energy_bins, fluxes, comment, norm)


@dispatch(TextIOBase)
def read_arb_fluxes(
    stream: TextIO,
) -> Fluxes:
    """Read arbitrary fluxes from stream.

    Args:
        stream: IO to read from

    Returns:
        Fluxes: the loaded
    """
    return read_fluxes(stream, define_arb_bins_and_fluxes)


@dispatch(Path)  # type: ignore[no-redef]
def read_arb_fluxes(path: Path) -> Fluxes:
    """Read arbitrary fluxes from Path.

    Args:
        path: Path to read from

    Returns:
        Fluxes: the loaded
    """
    with path.open() as stream:
        return read_arb_fluxes(stream)


@dispatch(str)  # type: ignore[no-redef]
def read_arb_fluxes(text: str) -> Fluxes:
    """Read arbitrary fluxes from text.

    Args:
        text: the text to read from

    Returns:
        Fluxes: the read
    """
    with StringIO(text) as stream:
        return read_fluxes(stream, define_arb_bins_and_fluxes)


@dispatch(TextIOBase)
def read_709_fluxes(
    stream: TextIO,
) -> Fluxes:
    """Read 709-group fluxes from stream.

    Args:
        stream: the stream to read from

    Returns:
        Fluxes: the read
    """
    return read_fluxes(stream, define_709_bins_and_fluxes)


@dispatch(Path)  # type: ignore[no-redef]
def read_709_fluxes(path: Path) -> Fluxes:
    """Read 709-group fluxes from path.

    Args:
        path: the path to read from

    Returns:
        Fluxes: the read
    """
    with path.open() as stream:
        return read_709_fluxes(stream)


@dispatch(str)  # type: ignore[no-redef]
def read_709_fluxes(text: str) -> Fluxes:
    """Read 709-group fluxes from text.

    Args:
        text: the text to read from

    Returns:
        Fluxes: the read
    """
    with StringIO(text) as stream:
        return read_709_fluxes(stream)


class FluxesDataSizeError(ValueError):
    """Base Fluxes read Exception."""

    def __str__(self) -> str:
        """Use an exception class __doc__ as an error msg.

        Returns:
            an exception class __doc__
        """
        return cast(str, self.__class__.__doc__)


class ArbitraryFluxesDataSizeError(FluxesDataSizeError):
    """The number of float values from arb_flux file should be odd."""


def define_arb_bins_and_fluxes(data: NDArrayFloat) -> tuple[NDArrayFloat, NDArrayFloat]:
    """Extract energy bins and values of arbitrary flux.

    Args:
        data: the array containing both bins and values

    Returns:
        Tuple: energy bins, values

    Raises:
        ArbitraryFluxesDataSizeError: if size of data is not odd.
    """
    sz = data.size
    if sz & 1 == 0:
        raise ArbitraryFluxesDataSizeError
    bins_end = sz // 2 + 1
    energy_bins = data[:bins_end]
    fluxes = data[bins_end:]
    energy_bins = energy_bins[::-1]
    fluxes = fluxes[::-1]
    return energy_bins, fluxes


class StandardFluxesDataSizeError(FluxesDataSizeError):
    """Invalid data for standard FISPACT 709-group fluxes."""


def define_709_bins_and_fluxes(data: NDArrayFloat) -> tuple[NDArrayFloat, NDArrayFloat]:
    """Strategy to define energy bins and values of standard 709 group flux.

    Args:
        data: the array containing values only

    Returns:
        Tuple: predefined energy bins, values

    Raises:
        StandardFluxesDataSizeError: if size of data is not 709.
    """
    if data.size != FISPACT_709_BINS_NUMBER:
        raise StandardFluxesDataSizeError
    return FISPACT_709_BINS, data[::-1]


def _print_bin_values(fluxes: Fluxes, fid: TextIO, max_columns: int = 5) -> None:
    """Print fluxes bins for FISPACT.

    Args:
        fluxes: to print bins from
        fid: a stream to print to
        max_columns: max columns in output
    """
    sequence = fluxes.fluxes[::-1]
    column = print_cols(sequence, fid, max_columns, fmt="{:.5e}")
    if column != 0:
        print(file=fid)
    print(fluxes.norm, file=fid)
    print(fluxes.comment, file=fid, end="")


class NotA709Error(FluxesDataSizeError):
    """Expected 709-group fluxes."""


def print_709_fluxes(fluxes: Fluxes, fid: TextIO, max_columns: int = 7) -> None:
    """Print standard 709-group fluxes.

    Args:
        fluxes: what to print
        fid: where
        max_columns: max columns in output

    Raises:
        NotA709Error: if not a valid 709 group "Fluxes" object is provided.
    """
    if not is_709_fluxes(fluxes):
        raise NotA709Error
    _print_bin_values(fluxes, fid, max_columns)


def print_arbitrary_fluxes(fluxes: Fluxes, fid: TextIO, max_columns: int = 5) -> None:
    """Print fluxes in FISPACT arbitrary flux format.

    Args:
        fluxes: what to print
        fid: output stream
        max_columns: max number of columns in a row
    """
    sequence = fluxes.energy_bins[::-1]
    column = print_cols(sequence, fid, max_columns, fmt="{:.6e}")
    if column != 0:
        print(file=fid)
    _print_bin_values(fluxes, fid, max_columns)



"""DAO implementation for DuckDB."""
from __future__ import annotations

from .implementation import DuckDBDAO, compute_optimal_row_group_size, write_parquet

__all__ = ["DuckDBDAO", "compute_optimal_row_group_size", "write_parquet"]



"""Nuclide specification in FISPACT JSON."""
from __future__ import annotations

from typing import Any

from dataclasses import dataclass

try:
    from scipy.constants import Avogadro
except ImportError:  # pragma: no cover
    Avogadro = 6.02214075999999987023872e23

from mckit_nuclides.elements import z
from mckit_nuclides.nuclides import get_nuclide_mass

__all__ = ["Avogadro", "Nuclide"]

FLOAT_ZERO = 0.0


@dataclass
class Nuclide:  # pylint: disable=too-many-instance-attributes
    """Nuclide properties from FISPACT JSON."""

    element: str
    isotope: int
    state: str = ""
    zai: int = 0  # Introduced in FISPACT-II v5: z*10000 + a*10 + (state? 1: 0)
    half_life: float = 0.0
    atoms: float = 0.0
    grams: float = 0.0
    activity: float = 0.0
    alpha_activity: float = 0.0  # FISPACT 5.0
    beta_activity: float = 0.0  # -/-
    gamma_activity: float = 0.0  # -/-
    heat: float = 0.0
    alpha_heat: float = 0.0
    beta_heat: float = 0.0
    gamma_heat: float = 0.0
    dose: float = 0.0
    ingestion: float = 0.0
    inhalation: float = 0.0

    def __post_init__(self) -> None:
        """Make the values consistent in data from old FISPACT."""
        _z = z(self.element)
        if self.zai == 0:
            self.zai = _z * 10000 + self.isotope * 10
            if self.state:
                self.zai += 1
        if self.atoms == FLOAT_ZERO and self.grams > FLOAT_ZERO:
            self.atoms = Avogadro * self.grams / get_nuclide_mass(_z, self.isotope)

    @property
    def a(self) -> int:
        """Synonym to mass number, isotope, A.

        Returns:
            A, mass number
        """
        return self.isotope

    @classmethod
    def from_json(cls, json_dict: dict[str, Any]) -> Nuclide:
        """Construct the Nuclide from JSON dictionary.

        Args:
            json_dict: information in json

        Returns:
            Nuclide: the Nuclide object
        """
        return cls(**json_dict)



"""Classes to read a FISPACT time step attributes from JSON."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from dataclasses import dataclass, field

import numpy as np

from xpypact.nuclide import Nuclide

if TYPE_CHECKING:
    from xpypact.utils.types import NDArrayFloat  # pragma: no cover

FLOAT_ZERO = 0.0


@dataclass
class DoseRate:
    """Dose rate attributes.

    Don't scale by mass, for point source mass is always 1g, for contact dose mass is meaningless.

    Attrs:
        type: "Plain source" for contact dose or "Point source"
        distance: specified for "point source", meters
        mass: mass for point source, always 1g
    """

    type: str = ""  # noqa: A003
    distance: float = 0.0
    mass: float = 0.0
    dose: float = 0.0

    def __post_init__(self) -> None:
        """Correct wrong value coming from FISPACT."""
        if self.mass == FLOAT_ZERO:
            # According to FISPACT manual (both v.4 and v.5 (p.63))
            # should be 1 gram always, but FISPACT shows 0 at the first step
            # and less, than 1 in the following steps. Fixing here.
            self.mass = 1.0e-3


@dataclass
class GammaSpectrum:
    """Data on gamma emission.

    Attrs:
        boundaries:
            Energy boundaries, MeV
        intensities:
            Gamma emission intensity.
    """

    boundaries: NDArrayFloat
    intensities: NDArrayFloat

    @classmethod
    def from_json(cls, json: dict[str, list[float]]) -> GammaSpectrum:
        """Construct GammaSpectrum instance from JSON dictionary.

        Args:
            json: dictionary

        Returns:
            The new GammaSpectrum instance with loaded boundaries and values.
        """
        return cls(
            boundaries=np.array(json["boundaries"], dtype=float),
            intensities=np.array(json["values"], dtype=float),
        )


@dataclass
class TimeStep:  # pylint: disable=too-many-instance-attributes
    """Time step attributes.

    All names must be the same as in FISPACT JSON file.
    """

    number: int = -1
    irradiation_time: float = 0.0
    cooling_time: float = 0.0
    duration: float = 0.0
    elapsed_time: float = 0.0
    flux: float = 0.0
    total_atoms: float = 0.0  # FISPACT 5.0
    total_activity: float = 0.0  # -/-
    alpha_activity: float = 0.0  # -/-
    beta_activity: float = 0.0  # -/-
    gamma_activity: float = 0.0  # -/-
    total_mass: float = 0.0  # -/-
    total_heat: float = 0.0
    alpha_heat: float = 0.0
    beta_heat: float = 0.0
    gamma_heat: float = 0.0
    ingestion_dose: float = 0.0
    inhalation_dose: float = 0.0
    dose_rate: DoseRate = field(default_factory=DoseRate)
    nuclides: list[Nuclide] = field(default_factory=list)
    gamma_spectrum: GammaSpectrum | None = None

    def __post_init__(self) -> None:
        """Correct data missed in FISPACT-4."""
        # workarounds for FISPACT v.4
        if self.total_mass == FLOAT_ZERO:
            self.total_mass = 1e-3 * sum(n.grams for n in self.nuclides)
        if self.total_atoms == FLOAT_ZERO:
            self.total_atoms = sum(n.atoms for n in self.nuclides)
        if self.total_activity == FLOAT_ZERO:
            self.total_activity = sum(n.activity for n in self.nuclides)
        if self.alpha_activity == FLOAT_ZERO:
            self.alpha_activity = sum(n.alpha_activity for n in self.nuclides)
        if self.beta_activity == FLOAT_ZERO:
            self.beta_activity = sum(n.beta_activity for n in self.nuclides)
        if self.gamma_activity == FLOAT_ZERO:
            self.gamma_activity = sum(n.gamma_activity for n in self.nuclides)

    @property
    def nuclides_mass(self) -> float:
        """Synonym for total mass.

        Returns:
            Total mass of the nuclides in kg.
        """
        return self.total_mass

    @property
    def is_cooling(self) -> bool:
        """Is the time step for cooling?

        Returns:
            Is the irradiation flux zero?
        """
        return self.flux == FLOAT_ZERO

    @classmethod
    def from_json(cls, json_dict: dict[str, Any]) -> TimeStep:
        """Construct TimeStep instance from JSON dictionary.

        Args:
            json_dict: source dictionary

        Returns:
            The new TimeStep instance.
        """
        json_dose_rate = json_dict.pop("dose_rate")
        dose_rate = DoseRate(**json_dose_rate)
        json_nuclides = json_dict.pop("nuclides")
        nuclides = [Nuclide.from_json(n) for n in json_nuclides] if json_nuclides else []
        json_gamma_spectrum = json_dict.pop("gamma_spectrum", None)
        if json_gamma_spectrum:
            gamma_spectrum = GammaSpectrum.from_json(json_gamma_spectrum)
        else:
            gamma_spectrum = None
        return cls(
            dose_rate=dose_rate,
            nuclides=nuclides,
            gamma_spectrum=gamma_spectrum,
            **json_dict,
        )



"""The `xpypact` package.

Wraps FISPACT workflow. Transforms FISPACT output to xarray datasets.
"""
from __future__ import annotations

from importlib import metadata as _meta
from importlib.metadata import PackageNotFoundError, version

from .data_arrays import from_json, scale_by_flux, scale_by_mass
from .data_frames import (
    get_gamma,
    get_nuclides,
    get_run_data,
    get_time_steps,
    get_timestep_nuclides,
)

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

__distribution__ = _meta.distribution(__name__)
__meta_data__ = __distribution__.metadata
__author__ = __meta_data__["Author"]
__author_email__ = __meta_data__["Author-email"]
__license__ = __meta_data__["License"]
__summary__ = __meta_data__["Summary"]
__copyright__ = f"Copyright 2021 {__author__}"

__all__ = [
    "__version__",
    "__distribution__",
    "__meta_data__",
    "__author__",
    "__author_email__",
    "__license__",
    "__summary__",
    "__copyright__",
    "get_run_data",
    "get_gamma",
    "get_nuclides",
    "get_timestep_nuclides",
    "get_time_steps",
    "from_json",
    "scale_by_flux",
    "scale_by_mass",
]



"""FISPACT run title data class."""
from __future__ import annotations

from dataclasses import dataclass


@dataclass
class RunData:
    """FISPACT run title data."""

    timestamp: str
    run_name: str
    flux_name: str

    @classmethod
    def from_json(cls, json_dict: dict[str, str]) -> RunData:
        """Construct RunData instance from JSON.

        Args:
            json_dict: source dictionary

        Returns:
            The loaded instance of RunData
        """
        return cls(**json_dict)



"""Interface to data access facilities."""
from __future__ import annotations

from typing import TYPE_CHECKING

from abc import ABC, abstractmethod

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr


class DataAccessInterface(ABC):
    """Abstract DAO for to save/load xpypact dataset.

    The subclasses implement methods to save/load xpypact dataset to/from
    database or filesystem.

    A caller is to provide connection or whatever resource to subclasses.
    """

    @abstractmethod
    def get_tables_info(self) -> pd.DataFrame:
        """Get information on tables in schema."""

    @abstractmethod
    def has_schema(self) -> bool:
        """Check if the schema is available in a database."""

    @abstractmethod
    def create_schema(self) -> None:
        """Create tables to store xpypact dataset.

        Retain existing tables.
        """

    @abstractmethod
    def drop_schema(self) -> None:
        """Drop our DB objects."""

    @abstractmethod
    def save(self, ds: xr.Dataset, material_id=1, case_id="") -> None:
        """Save xpypact dataset to database.

        Args:
            ds: xpypact dataset to save
            material_id: additional key to distinguish multiple FISPACT run
            case_id: second additional key
        """

    @abstractmethod
    def load_rundata(self) -> pd.DataFrame:
        """Load FISPACT run data as table.

        Returns:
            FISPACT run data
        """

    @abstractmethod
    def load_nuclides(self) -> pd.DataFrame:
        """Load nuclide table.

        Returns:
            time nuclide
        """

    @abstractmethod
    def load_time_steps(self) -> pd.DataFrame:
        """Load time step table.

        Returns:
            time step table
        """

    @abstractmethod
    def load_time_step_nuclides(self) -> pd.DataFrame:
        """Load time step x nuclides table.

        Returns:
            time step x nuclides table
        """

    @abstractmethod
    def load_gamma(self, time_step_number: int | None = None) -> pd.DataFrame:
        """Load time step x gamma table.

        Args:
            time_step_number: filter for time_step_number

        Returns:
            time step x gamma table
        """
